/* $Id: sfsagent.C,v 1.27 1999/04/22 03:08:17 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "arpc.h"
#include "sfsagent.h"
#include "agentmisc.h"
#include "crypt.h"

ptr<aclnt> cagent;

struct key {
  ptr<rabin_priv> k;
  u_int32_t expire;
  str comment;
  list_entry<key> link;
};

struct sfsagent {
  bool main;
  str name;
  ptr<axprt> x;
  ptr<asrv> as;
  ptr<asrv> cs;

  sfsagent (str name, int fd, bool main = false);
  void dispatch (svccb *sbp);
  void ctldisp (svccb *sbp);
  void agentdisp (svccb *sbp);
  void authinit (svccb *sbp);
};

static sfsagent *a;
list<key, &key::link> keys;
static int startproc = AGENT_START;

#if 0
static void
nullccb_fn (clnt_stat)
{
}
static aclnt_cb nullccb (wrap (nullccb_fn));
static int nullint;
#endif

static void
startcb (int32_t *res, clnt_stat stat)
{
  if (stat)
    fatal ("cannot register agent: %s\n", clnt_sperrno (stat));
  if (*res)
    fatal ("cannot register agent: %s\n", strerror (*res));
  delete res;
  startproc = AGENT_START;
}

static void
symlink_cb (svccb *sbp, clnt_stat err)
{
  sbp->reply (NULL);
}

static key *
klookup (const bigint &n)
{
  key *k;
  for (k = keys.first; k; k = k->link.next)
    if (k->k->n == n)
      return k;
  return NULL;
}

static key *
keynum (u_int n)
{
  key *k = keys.first;
  while (k && n--)
    k = k->link.next;
  return k;
}

sfsagent::sfsagent (str n, int fd, bool m)
  : main (m), name (n), x (axprt_stream::alloc (fd)),
    as (asrv::alloc (x, agentcb_prog_1, wrap (this, &sfsagent::dispatch)))
{
  if (!main)
    cs = asrv::alloc (x, agentctl_prog_1, wrap (this, &sfsagent::dispatch));
}

void
sfsagent::authinit (svccb *sbp)
{
  sfsagent_authinit_arg *aa = sbp->template getarg<sfsagent_authinit_arg> ();
  key *k = keynum (aa->ntries);
  if (!k || aa->authinfo.type != SFS_AUTHINFO) {
    sbp->replyref (sfsagent_auth_res (false));
    return;
  }

  sfsagent_auth_res res (true);
  sfs_autharg ar (SFS_AUTHREQ);
  sfs_signed_authreq sar;
  str rawsar;

  ar.req->usrkey = k->k->n;
  sar.type = SFS_SIGNED_AUTHREQ;
  sar.seqno = aa->seqno;
  bzero (sar.usrinfo.base (), sar.usrinfo.size ());

  if (!sha1_hashxdr (sar.authid.base (), aa->authinfo)
      || !(rawsar = xdr2str (sar))
      || !(ar.req->signed_req = k->k->sign_r (rawsar),
	   xdr2bytes (*res.certificate, ar))) {
    warn ("sfsagent::authinit: xdr failed\n");
    res.set_authenticate (false);
  }

  warn << aa->requestor << ": " << aa->authinfo.name << ":"
       << armor32 (str (aa->authinfo.hostid.base (),
			aa->authinfo.hostid.size ()))
       << " (" << implicit_cast<int> (aa->authinfo.service) << ")\n";

  sbp->replyref (res);
}

void
sfsagent::dispatch (svccb *sbp)
{
  if (!sbp) {
    if (main)
      fatal << "EOF from " << name << "\n";
    // warn << "EOF from " << name << "\n";
    delete this;
    return;
  }

  if (sbp->prog () == AGENTCB_PROG && sbp->vers () == AGENTCB_VERS)
    agentdisp (sbp);
  else if (sbp->prog () == AGENTCTL_PROG && sbp->vers () == AGENTCTL_VERS)
    ctldisp (sbp);
  else
    panic ("invalid prog/vers\n");
}

void
sfsagent::agentdisp (svccb *sbp)
{
  switch (sbp->proc ()) {
  case AGENTCB_NULL:
    sbp->reply (NULL);
    break;
  case AGENTCB_AUTHINIT:
    authinit (sbp);
    break;
  case AGENTCB_AUTHMORE:
    sbp->replyref (sfsagent_auth_res (false));
    break;
  default:
    warn ("invalid procno %d\n", sbp->proc ());
    sbp->reject (PROC_UNAVAIL);
  }
}

void
sfsagent::ctldisp (svccb *sbp)
{
  switch (sbp->proc ()) {
  case AGENTCTL_NULL:
    sbp->reply (NULL);
    break;
  case AGENTCTL_ADDKEY:
    {
      sfs_addkey_arg *aa = sbp->template getarg<sfs_addkey_arg> ();
      key *nk = New key;
      nk->k = rabin_priv::make (aa->p, aa->q);
      if (klookup (nk->k->n)) {
	delete nk;
	sbp->replyref (false);
      }
      else {
	nk->expire = aa->expire;
	nk->comment = aa->comment;
	keys.insert_head (nk);
	int *res = New int;
	cagent->call (startproc, NULL, res, wrap (startcb, res));
	sbp->replyref (true);
      }
      break;
    }
  case AGENTCTL_REMKEY:
    {
      sfs_remkey_arg *ra = sbp->template getarg<sfs_remkey_arg> ();
      key *k = klookup (*ra);
      if (k) {
	keys.remove (k);
	delete k;
	int *res = New int;
	cagent->call (startproc, NULL, res, wrap (startcb, res));
	sbp->replyref (true);
      }
      else
	sbp->replyref (false);
      break;
    }
  case AGENTCTL_REMALL:
    {
      for (key *k = keys.first, *nk; k; k = nk) {
	nk = k->link.next;
	keys.remove (k);
	delete k;
      }
      int *res = New int;
      cagent->call (startproc, NULL, res, wrap (startcb, res));
      sbp->replyref (NULL);
      break;
    }
  case AGENTCTL_DUMP:
    {
      sfs_keylist kl;
      sfs_keylist *klp = &kl;
      for (key *k = keys.first; k; k = k->link.next) {
	(*klp).alloc ();
	(*klp)->key = k->k->n;
	(*klp)->expire = k->expire;
	(*klp)->comment = k->comment;
	klp = &(*klp)->next;
      }
      sbp->reply (&kl);
      break;
    }
  case AGENTCTL_SYMLINK:
    {
      cagent->call (AGENT_SYMLINK, sbp->getvoidarg (), NULL, 
		    wrap (symlink_cb, sbp));
      break;
    }
  default:
    warn ("invalid procno %d\n", sbp->proc ());
    sbp->reject (PROC_UNAVAIL);
  }
}

static void
newctl (int fd)
{
  if (fd > 0)
    vNew sfsagent ("sock", fd);
}

static void
agent_daemonize ()
{
  if (!keys.first)
    fatal << "No keys with which to authenticate user!\n";

  switch (fork ()) {
  case -1:
    fatal ("fork: %m\n");
  case 0:
    break;
  default:
    exit (0);
  }
  if (setsid () < 0)
    fatal ("setsid: %m\n");
  chdir ("/");
}

static void
start_listening (int status)
{
  if (status) {
    int err = WIFEXITED (status) ? WEXITSTATUS (status) : EFAULT;
    fatal << agentsock << ": " << strerror (err) << "\n";
  }
  agent_daemonize ();
  sfs_unixserv (agentsock, wrap (newctl));
}

static void
prep_socket (bool Sset, int status)
{
  if (status)
    fatal << "sfskey failed\n";
  if (!agentsock) {
    agent_daemonize ();
    return;
  }

  pid_t pid = afork ();
  if (pid == -1)
    fatal ("fork: %m\n");
  if (!pid) {
    /* Trigger any necessary authentication here, so that we don't
     * deadlock in the parent process. */
    if (!Sset) {
      usrdir_make ();
      usrdir_agentsock ();
    }
    struct stat sb;
    if (!stat (agentsock, &sb)) {
      if (!S_ISSOCK (sb.st_mode))
	_exit (EINVAL);
      if (startproc == AGENT_KILLSTART)
	unlink (agentsock);
      unlink (agentsock);
      _exit (0);
    }
    if (errno == ENOENT)
      _exit (0);
    _exit (errno);
  }
  chldcb (pid, wrap (start_listening));
}

static void
startagent (int argc, char **argv)
{
  bool Sset = false;

  vector<str> av;
  // av.push_back (find_program ("sfskey", argv[0]));
  av.push_back (find_program ("sfskey"));// XXX
  if (!av[0])
    fatal ("cannot find sfskey\n");
  av.push_back ("add");
  av.push_back ("-S");
  av.push_back (NULL);
  for (int i = 1; i < argc; i++) {
    if (argv[i][0] == '-' && argv[i][1] == 'S') {
      Sset = true;
      if (argv[i][2])
	agentsock = argv[i] + 2;
      else if (++i < argc)
	agentsock = argv[i];
      else
	fatal << "missing argument for -S\n";
      continue;
    }
    if (argv[i][0] == '-' && argv[i][1] == 'k' && argv[i][2] == '\0') {
      startproc = AGENT_KILLSTART;
      continue;
    }
    av.push_back (argv[i]);
  }

  int fds[2];
  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0)
    fatal ("socketpair: %m\n");
  av[3] = strbuf ("-%d", fds[1]);

  vector<char *> avcp;
  for (str *sp = av.base (); sp < av.lim (); sp++)
    avcp.push_back (const_cast<char *> (sp->cstr ()));
  avcp.push_back (NULL);

  pid_t pid = afork ();
  if (pid == -1)
    fatal ("fork: %m\n");
  if (!pid) {
    close (fds[0]);
    execv (avcp[0], avcp.base ());
    fatal ("%s: %m\n", avcp[0]);
  }
  close (fds[1]);
  vNew sfsagent ("sfskey", fds[0]);

  chldcb (pid, wrap (prep_socket, Sset));
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);

  usrdir_agentsock ();
  random_init ();

  /* This makes ptrace fail on some OS's (a good thing, when we hold
   * private keys). */
  if (!runinplace)
    setuid (getuid ());

  int fd = suidgetfd_required ("agent");
  a = New sfsagent ("sfscd", fd, true);
  cagent = aclnt::alloc (a->x, agent_prog_1);

  startagent (argc, argv);

  amain ();
}
