/* $Id: sfskeyupdate.C,v 1.18 1999/05/07 09:06:02 dm Exp $ */

/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfskey.h"

void
keyupdate::client_keygen ()
{
  client_key = New refcounted<rabin_priv> (rabin_keygen (pubkeybits));
  timecb (time (NULL) + 3600, wrap (this, &keyupdate::client_keygen));
}

void
keyupdate::finish (int err)
{
  (*cb) (err);
  delete this;
}

void
keyupdate::doupdate_cb (sfsauth_updateres *_resp, clnt_stat err)
{
  auto_ptr<sfsauth_updateres> resp (_resp);

  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  str user (getlogin ());

  switch (*resp) {
  case SFSAUTH_UPDATE_OK:
    {
      warn << "authserv: Update successfull...installing new key:\n";
      warnx << user << ":0x" << newkey->modulus ().getstr (16) << "\n";
      finish (0);
      return;
    }
  case SFSAUTH_UPDATE_BADSIGNATURE:
    warn << "authserv: Update failed--bad signature.\n";
    finish (1);
    return;
  case SFSAUTH_UPDATE_PROTOERR:
    warn << "authserv: Update failed--protocol error!\n";
    finish (1);
    return;
  case SFSAUTH_UPDATE_NOTTHERE:
    warn << "authserv: Update failed--following user is not registered:\n";
    warnx << user << ":0x" << oldkey->modulus ().getstr (16) << "\n";
    finish (1);
    return;
  case SFSAUTH_UPDATE_BADAUTHID:
    warn << "authserv: Update failed--bad authid.\n";
    finish (1);
    return;
  case SFSAUTH_UPDATE_DENYFILE:
    warn << "authserv: Update failed--user in deny file.\n";
    finish (1);
    return;
  case SFSAUTH_UPDATE_FAILED:
    warn << "authserv: Update failed--could not update user!\n";
    finish (1);
    return;
  default:
    warn << "Unknown response from authserv\n";
    finish (1);
    return;
  }
}

void
keyupdate::gotsrpparams (sfsauth_srpinfo *resp, sfs_hash sessid, 
			 clnt_stat err)
{
  if (err) {
    fatal << "authserv: " << err << "\n";
  }

  sfsauth_updatearg arg;
  arg.updatereq.type = SFS_AUTHUPDATE;
  arg.updatereq.oldkey = oldkey->modulus ();
  arg.updatereq.newkey = newkey->modulus ();

  /* Get private key parameters, if needed */
  if (opt_sec) {
    srp srpclnt;

    if (!srpfile) {
      if (!srpclnt.set_params (bigint (resp->prime, 16),
			       bigint (resp->generator, 16)))
	fatal << "SRP parameters N and g failed to check out!\n";
    }
    else {
      srpclnt.set_params (bigint (resp->prime, 16),
			  bigint (resp->generator, 16), false);
    }

    str salt = pw_gensalt (SK_SRP_EKSBF_COST, authhostname);

    eksblowfish eksb;
    bigint x = pw_getint (newpwd, salt, srpclnt.N.nbits () - 1, &eksb);
    if (!x)
      fatal << "SRP modulus too short\n";

    arg.updatereq.srpinfo.alloc ();
    arg.updatereq.srpinfo->prime = resp->prime;
    arg.updatereq.srpinfo->generator = resp->generator;
    arg.updatereq.srpinfo->salt = salt;
    arg.updatereq.srpinfo->verifier = srpclnt.calc_verifier (x).getstr (16);
    arg.updatereq.srpinfo->seckey = export_rabin_encrypt_sec (*newkey, &eksb);
    if (!arg.updatereq.srpinfo->seckey)
      fatal << "Could not encrypt secret key\n";
  }

  /* Build the authid structure */
  sfs_authinfo authinfo;
  authinfo.type = SFS_AUTHINFO;
  authinfo.service = SFS_AUTHSERV;
  authinfo.name = authhostname;
  authinfo.hostid = authhostid;
  authinfo.sessid = sessid;

  if (!sha1_hashxdr (arg.updatereq.authid.base (), authinfo)) {
    fatal << "Could not hash authinfo into authid.\n";
  }

  str rawupdatereq;
  if (!(rawupdatereq = xdr2str (arg.updatereq))) {
    fatal << "Could not translate updatereq into a string.\n";
  }
  arg.signature = oldkey->sign (rawupdatereq);

  delete resp;

  sfsauth_updateres *resp2 = New sfsauth_updateres;
  authclnt->call (SFSAUTHPROC_UPDATE, &arg, resp2,
		  wrap (this, &keyupdate::doupdate_cb, resp2));
}

void
keyupdate::cryptres (const sfs_hash *sessidp)
{
  if (!sessidp) {
    warn << "authserv: unable to encrypt connection" << "\n";
    finish (EIO);
    return;
  }

  sfsauth_srpinfo *resp = New sfsauth_srpinfo;

  if (opt_sec) {
    warn << "Checking SRP parameters.  This may take a few seconds.\n";

    if (!newpwd)
      fatal << "Private key has no password--can't install on authserv\n";

    if (!srpfile)
      authclnt->call (SFSAUTHPROC_SRP_GETGLOBALPARAMS, NULL, resp,
		      wrap (this, &keyupdate::gotsrpparams, resp, *sessidp));
    else {
      ptr<srp> s = import_srp_params (file2str (srpfile));
      if (!s)
	fatal << "Could not load SRP parameters from " << srpfile << "\n";
      resp->prime = s->N.getstr (16);
      resp->generator = s->g.getstr (16);
      gotsrpparams (resp, *sessidp, RPC_SUCCESS);
    }
  }
  else {
    gotsrpparams (resp, *sessidp, RPC_SUCCESS);
  }
}

void
keyupdate::sfsconres (sfs_connectres *resp, clnt_stat err)
{
  if (err) {
    warn << "authserv: " << err << "\n";
    finish (EIO);
  }
  else if (resp->status) {
    warn << "authserv: " << resp->status << "\n";
    finish (resp->status);
  }
  else if (!sfs_ckhostid (&authhostid, resp->reply->host)) {
    warn << "authserv: hostid from server doesn't match given hostid\n";
    finish (ENOENT);
  }
  else {
    sfs_client_crypt (sfsclnt, client_key, resp->reply->host.pubkey,
 		      wrap (this, &keyupdate::cryptres));
  }

  delete resp;
}

void
keyupdate::connected (int fd)
{
  if (fd < 0) {
    warn << "Couldn't not connect to authserv " << strerror (errno) << "\n";
    finish (errno);
  }
  else {
    authclnt = aclnt::alloc (authxprt = axprt_crypt::alloc (fd),
			     sfsauth_program_1);
    sfsclnt = aclnt::alloc (authxprt, sfs_program_1);

    sfs_connectarg arg;
    arg.release = SFS_RELEASE;
    arg.service = SFS_AUTHSERV;
    arg.name = authhostname;
    arg.hostid = authhostid;

    sfs_connectres *resp = New sfs_connectres;
    sfsclnt->call (SFSPROC_CONNECT, &arg, resp,
		   wrap (this, &keyupdate::sfsconres, resp));
  }
}

void
keyupdate::connect ()
{
  if (!authserv)
    fatal << "No authserv specified; use; use hostname:hostid format\n";
  if (!sfs_parsepath (authserv, &authhostname, &authhostid)) {
    fatal << "Could not create hostid and hostname from host info.\n";
  }

#if 0
  int fd = suidgetfd ("authserv");
#else
  int fd = -1;
#endif

  /* Do the RPC to the authserv */
  if (fd < 0) {
    tcpconnect (authhostname, sfs_port,	wrap (this, &keyupdate::connected));
  }
  else {
    connected (fd);
  }

}

void
keyupdate::gothostid (str path)
{
  if (!path) {
    warn << "Unable to fetch hostid from server.\n";
    finish (1);
    return;
  }

  authserv = path;
  connect ();
}

void
keyupdate::gotpwdkey2 (str pwd)
{
  newpwd = pwd;
  newkey = import_rabin_priv (newkey_asc, newpwd);

  if (!newkey)
    fatal << "could not decode secret key\n";

  if (oldkeyfile == newkeyfile)
    oldkey = newkey;

  if (authserv)
    connect ();
  else
    sfs_gethostid (NULL, SFS_AUTHSERV, wrap (this, &keyupdate::gothostid));
}

void
keyupdate::gotpwdkey1 (ptr <rabin_priv> sk)
{
  if (oldkeyfile != newkeyfile) {
    if (!sk)
      fatal ("could not decode secret key\n");
    oldkey = sk;
  }

  newkey_asc = file2wstr (newkeyfile);
  if (!newkey_asc)
    fatal ("%s: %m\n", newkeyfile.cstr ());

  if (import_rabin_priv_need_pwd (newkey_asc)) {
    if (!getkbdpwd ("Passphrase for " << newkeyfile << ": ",
		    &rnd_input, wrap (this, &keyupdate::gotpwdkey2)))
      fatal ("stdin must be a tty.\n");
  }
  else {
    gotpwdkey2 (NULL);
  }
}

keyupdate::keyupdate (str oldkf, str newkf, str srpf, 
		      bool sec, str as, cb_t cb)
  : oldkeyfile (oldkf), newkeyfile (newkf), srpfile (srpf), 
    opt_sec (sec), authserv (as), cb (cb)
{
  client_keygen ();

  if (oldkeyfile == newkeyfile) {
    gotpwdkey1 (NULL);
  }
  else {
    str sk (file2wstr (oldkeyfile));
    if (!sk)
      fatal ("%s: %m\n", oldkeyfile.cstr ());

    import_rabin_priv_askpwd (sk, "Passphrase for " << oldkeyfile << ": ",
			      wrap (this, &keyupdate::gotpwdkey1));
  }
}
