/* $Id: authclient.C,v 1.31 1999/05/28 21:50:27 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "authserv.h"
#include "rxx.h"
#if HAVE_GETSPNAM
#include "shadow.h"
#endif /* HAVE_GETSPNAM */

ptr<srp> srpglobal;
ptr<rabin_priv> authsrvkey;
sfs_hostinfo hinfo;
sfs_hash hostid;

extern "C" char *crypt (const char *, const char *);

authclient::authclient (ptr<axprt_stream> _x, const authunix_parms *aup)
  : x (_x)
{
  if ((unixauth = aup))
    uid = aup->aup_uid;

  authsrv = asrv::alloc (x, sfsauth_program_1,
			 wrap (this, &authclient::dispatch));
  sfssrv = asrv::alloc (x, sfs_program_1,
			wrap (this, &authclient::dispatch));

  authid_valid = false;
}

void
authclient::dologin (svccb *sbp)
{
  sfs_loginarg *larg = sbp->template getarg<sfs_loginarg> ();

  sfsauth_loginres res (SFSLOGIN_BAD);
  sfs_autharg aarg;

  if (bytes2xdr (aarg, larg->certificate)) {
    switch (aarg.type) {
    case SFS_NOAUTH:
      res.set_status (SFSLOGIN_OK);
      bzero (&res.resok->authid, sizeof (res.resok->authid));
      res.resok->seqno = larg->seqno;
      res.resok->cred.set_type (SFS_NOCRED);
      break;
    case SFS_AUTHREQ:
      {
	// Check the signature of signed_req with usrkey
	sfs_signed_authreq authreq;
	rabin_pub usrkey (aarg.req->usrkey);
	str msg = usrkey.verify_r (aarg.req->signed_req, sizeof (authreq));
	authentry ae;
	ae.pubkey = "0x" << usrkey.n.getstr (16);

	if (msg && str2xdr (authreq, msg)
	    && authreq.type == SFS_SIGNED_AUTHREQ
	    && authreq.seqno == larg->seqno
	    && authlookup (&ae)) {
	  res.set_status (SFSLOGIN_OK);
	  res.resok->cred.set_type (SFS_UNIXCRED);
	  if (unixlookup (&ae, res.resok->cred.unixcred.addr ())) {
	    res.resok->authid = authreq.authid;
	    res.resok->seqno = authreq.seqno;
	  }
	  else
	    res.set_status (SFSLOGIN_BAD);
	}
	break;
      }
    }
  }
  sbp->reply (&res);
}

void
authclient::doregister (svccb *sbp)
{
  sfsauth_registerarg *arg = sbp->template getarg<sfsauth_registerarg> ();

  /* Make sure that we talking over the unix domain socket? */
  if (!unixauth) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_NOTSOCK));
    return;
  }

  struct passwd *pe;
  if ((pe = getpwnam (arg->username)) == NULL) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_BADUSERNAME));
    return;
  }

#if HAVE_GETSPNAM
  struct spwd *spe;
  if ((spe = getspnam (arg->username)) == NULL) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_BADUSERNAME));
    return;
  }
#endif /* HAVE_GETSPNAM */

  /* XXX - In FreeBSD, pw_uid is an int.  Go figure.  This isn't a
   * correct fix.  Maybe uid should be a u_int32_t and pw_uid cast to
   * that. */
  if (implicit_cast<uid_t> (pe->pw_uid) != uid) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_WRONGUID));
    return;
  }

  if (uid == 0) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_DENYROOT));
    return;
  }

  if (!validshell (pe->pw_shell)) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_BADSHELL));
    return;
  }

#if HAVE_GETSPNAM
  if (strcmp (spe->sp_pwdp, crypt (arg->password.cstr (), spe->sp_pwdp))) {
#else /* !HAVE_GETSPNAM */
  if (strcmp (pe->pw_passwd, crypt (arg->password.cstr (), pe->pw_passwd))) {
#endif /* !HAVE_GETSPNAM */
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_BADPASSWORD));
    return;
  }

  authentry ae;
  ae.user = arg->username;

  if (authlookup (&ae)) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_USEREXISTS));
    return;
  }

  if (indenyfile (arg->username)) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_DENYFILE));
    return;
  }

  rabin_pub usrkey (arg->pubkey);
  ae.pubkey = "0x" << usrkey.n.getstr (16);
  ae.user = arg->username;

  if (arg->srpinfo) {
    ae.srpinfo = arg->srpinfo->prime
      << "," << arg->srpinfo->generator
      << "," << arg->srpinfo->salt
      << "," << arg->srpinfo->verifier
      << "," << arg->srpinfo->seckey;
  }
  else
    ae.srpinfo = "";

  if (!authadd (&ae)) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_FAILED));
    return;
  }

  sbp->replyref (sfsauth_registerres (SFSAUTH_REGISTER_OK));
  return;
}

void
authclient::doupdate (svccb *sbp)
{
  sfsauth_updatearg *arg = sbp->template getarg<sfsauth_updatearg> ();

  // Check the signature of signed_req with usrkey
  rabin_pub oldkey (arg->updatereq.oldkey);
  str rawupdatereq = xdr2str (arg->updatereq);

  if (!oldkey.verify (rawupdatereq, arg->signature)) {
    sbp->replyref (sfsauth_updateres (SFSAUTH_UPDATE_BADSIGNATURE));
    return;
  }

  if (arg->updatereq.type != SFS_AUTHUPDATE) {
    sbp->replyref (sfsauth_updateres (SFSAUTH_UPDATE_PROTOERR));
    return;
  }

  if (!authid_valid || arg->updatereq.authid != authid) {
    sbp->replyref (sfsauth_updateres (SFSAUTH_UPDATE_BADAUTHID));
    return;
  }

  authentry olde;
  olde.pubkey = "0x" << oldkey.n.getstr (16);

  if (!authlookup (&olde)) {
    sbp->replyref (sfsauth_updateres (SFSAUTH_UPDATE_NOTTHERE));
    return;
  }

  if (indenyfile (olde.user)) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_UPDATE_DENYFILE));
    return;
  }

  authentry newe;
  rabin_pub newkey (arg->updatereq.newkey);
  newe.pubkey = "0x" << newkey.n.getstr (16);
  newe.user = olde.user;

  if (arg->updatereq.srpinfo) {
    newe.srpinfo = arg->updatereq.srpinfo->prime
      << "," << arg->updatereq.srpinfo->generator
      << "," << arg->updatereq.srpinfo->salt
      << "," << arg->updatereq.srpinfo->verifier
      << "," << arg->updatereq.srpinfo->seckey;
  }
  else
    newe.srpinfo = "";

  if (!authupdate (&olde, &newe)) {
    sbp->replyref (sfsauth_registerres (SFSAUTH_UPDATE_FAILED));
    return;
  }

  sbp->replyref (sfsauth_updateres (SFSAUTH_UPDATE_OK));
}

void
authclient::sendglobalparams (svccb *sbp)
{
  sfsauth_srpinfo res;

  if (!srpglobal) {
    res.prime = "";
    res.generator = "";
  }
  else {
    res.prime = srpglobal->N.getstr (16);
    res.generator = srpglobal->g.getstr (16);
  }
  res.salt = "";
  res.verifier = "";
  res.seckey = "";

  sbp->replyref (res);
}

#define A64STR "[A-Za-z0-9+/]+={0,2}"
#define HEXSTR "[0-9a-f]+"
#define AENTRY "^(" HEXSTR "),(" HEXSTR "),(\\S+),(" HEXSTR "),(" A64STR ")$"

void
authclient::dosrp0 (svccb *sbp)
{
  sfsauth_srp0arg *arg = sbp->template getarg<sfsauth_srp0arg> ();

#if 0
  warn << "username = " << arg->username << "\n";
#endif

  authentry e;
  e.user = arg->username;

  if (!authlookup (&e)) {
    sbp->replyref (sfsauth_srp0res (SFSAUTH_SRP_NOTTHERE));
    return;
  }

  rxx r (rxx (AENTRY));
  if (!r.match (e.srpinfo)) {
    warn << "Can't parse srpinfo.\n";
    sbp->replyref (sfsauth_srp0res (SFSAUTH_SRP_BADSRPINFO));
    return;
  }

  if (!srpserv.set_params (bigint (r[1], 16), bigint (r[2], 16), false)) {
    sbp->replyref (sfsauth_srp0res (SFSAUTH_SRP_BADSRPINFO));
    return;
  }

  if (!srpserv.phase0serv (bigint (r[4], 16), r[3], e.user)) {
    sbp->replyref (sfsauth_srp0res (SFSAUTH_SRP_PHASE0));
    return;
  }

  /* Keep secret information (key) around for later */
  seckey = str2wstr (r[5]);

  sfsauth_srp0res res (SFSAUTH_SRP_OK);
  res.resok->prime = r[1];
  res.resok->generator = r[2];

  sbp->replyref (res);
}

void
authclient::dosrp1 (svccb *sbp)
{
  sfsauth_srp1arg *arg = sbp->template getarg<sfsauth_srp1arg> ();

#if 0
  warn << "A = " << arg->A << "\n";
#endif

  if (!srpserv.phase1serv (bigint (arg->A, 16))) {
    sbp->replyref (sfsauth_srp1res (SFSAUTH_SRP_PHASE1));
    return;
  }

  sfsauth_srp1res res (SFSAUTH_SRP_OK);
  res.resok->salt = srpserv.salt;
  res.resok->B = srpserv.B.getstr (16);
  res.resok->u = srpserv.u.getstr (16);

  sbp->replyref (res);
}

void
authclient::dosrp2 (svccb *sbp)
{
  sfsauth_srp2arg *arg = sbp->template getarg<sfsauth_srp2arg> ();

  if (!srpserv.phase2serv (reinterpret_cast<u_char *> 
			   (arg->clntproof.base ()))) {
    sbp->replyref (sfsauth_srp2res (SFSAUTH_SRP_PHASE2));
    return;
  }

  sfsauth_srp2res res (SFSAUTH_SRP_OK);
  memcpy (res.resok->servproof.base (), srpserv.proof, 
	  res.resok->servproof.size ());

  sbp->replyref (res);

#if 0
  warn << "Session Key = " << srpserv.S << "\n";
#endif

  sfsauth_server_crypt (authsrv, srpserv.S);
}

void
authclient::dofetch (svccb *sbp)
{
  sfsauth_fetchres res (SFSAUTH_FETCH_OK);
  res.resok->seckey = seckey;
  res.resok->hostid = hostid;

  sbp->replyref (res);
}

void
authclient::dispatch (svccb *sbp)
{
  if (!sbp) {
    delete this;
    return;
  }

  if (sbp->prog () == SFS_PROGRAM) {
    switch (sbp->proc ()) {
    case SFSPROC_NULL:
      sbp->reply (NULL);
      return;
    case SFSPROC_CONNECT:
      {
	sfs_connectres res (SFS_OK);
	res.reply->release = SFS_RELEASE;
	res.reply->host = hinfo;
	res.reply->prog = SFSAUTH_PROGRAM;
	res.reply->vers = SFSAUTH_VERSION;
	bzero (&res.reply->charge, sizeof (sfs_hashcharge));
	sbp->reply (&res);
	return;
      }
    case SFSPROC_ENCRYPT:
      {
	sfs_hash sessid;

	sfs_server_crypt (sbp, authsrvkey, &sessid);
	sfs_get_authid (&authid, SFS_AUTHSERV, hinfo.hostname,
			&hostid, &sessid);
	authid_valid = true;
      }
    }
  }
  else {
    switch (sbp->proc ()) {
    case SFSAUTHPROC_NULL:
      sbp->reply (NULL);
      break;
    case SFSAUTHPROC_LOGIN:
      dologin (sbp);
      break;
    case SFSAUTHPROC_REGISTER:
      doregister (sbp);
      break;
    case SFSAUTHPROC_UPDATE:
      doupdate (sbp);
      break;
    case SFSAUTHPROC_SRP_GETGLOBALPARAMS:
      sendglobalparams (sbp);
      break;
    case SFSAUTHPROC_SRP_PHASE0:
      dosrp0 (sbp);
      break;
    case SFSAUTHPROC_SRP_PHASE1:
      dosrp1 (sbp);
      break;
    case SFSAUTHPROC_SRP_PHASE2:
      dosrp2 (sbp);
      break;
    case SFSAUTHPROC_FETCH:
      dofetch (sbp);
      break;
    default:
      sbp->reject (PROC_UNAVAIL);
      break;
    }
  }
}
