/* $Id: authdb.C,v 1.5 1999/05/28 21:50:27 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "authdb.h"
#include "rxx.h"

vector<userfile> userfiles;
ptr<userfile> shareduserfile;

class authfile {
  str filename;
  int fd;
  int lineno;
  suio in;

  void reset ();
public:
  authfile () : fd (-1) {}
  ~authfile () { close (fd); }

  bool openfile (str path, int flags, int perms);
  void closefile () { reset (); }
  bool exists (str path);
  bool getentry (authentry *);
};

static str
authentry2str (authentry *e)
{
  str s = e->user << ":" << e->pubkey << ":"
                  << (e->srpinfo ? e->srpinfo.cstr () : "") << "\n";
  return s;
}

void
authfile::reset ()
{
  filename = NULL;
  if (fd >= 0)
    close (fd);
  fd = -1;
  lineno = 0;
  in.clear ();
}

bool
authfile::openfile (str path, int flags, int perms = 0600)
{
  reset ();
  fd = open (path, flags, perms);
  if (fd < 0) {
    warn << path << ": " << strerror (errno) << "\n";
    return false;
  }
  filename = path;
  return true;
}

bool
authfile::exists (str file)
{
  if (!access (file, F_OK))
    return true;
  if (errno != ENOENT)
    fatal << file << ": " << strerror (errno) << "\n";
  return false;
}

static rxx authrx ("^([^:]+):([^:]+):([^:]+)?$");
bool
authfile::getentry (authentry *e)
{
  str line;

  do {
    while (!(line = suio_getline (&in)))
      if (in.input (fd, 8192) <= 0) {
	if (in.resid ()) {
	  warn << filename << ":" << lineno + 1 << ": incomplete last line\n";
	  in.clear ();
	}
	return false;
      }
    lineno++;
    authrx.match (line);
    if (!authrx.success ())
      warn << filename << ":" << lineno << ": malformed line ignored\n";
  } while (!authrx.success ());

  e->user = authrx[1];
  e->pubkey = authrx[2];
  e->srpinfo = authrx[3];
  if (!e->srpinfo)
    e->srpinfo = "";

  return true;
}

bool
authlookup (authentry *e)
{
  authfile af;
  bool bykey;

  if (e->pubkey)
    bykey = true;
  else if (e->user)
    bykey = false;
  else
    return false;

  for (size_t i = 0; i < userfiles.size (); i++) {
    if (!af.openfile (userfiles[i].path, O_RDONLY))
      continue;
    authentry ae;
    while (af.getentry (&ae))
      if ((bykey && e->pubkey == ae.pubkey)
	  || (!bykey && e->user == ae.user)) {
	*e = ae;
	return true;
      }
  }
  return false;
}

static bool
authadd_core (authentry *e, userfile file, int perm)
{
  authfile af;
  str astr;

  astr = file2str (file.path);
  if (!astr) {
    warn << "authadd: error reading " << file.path 
	 << ".  Make sure it exists.\n";
    return false;
  }
  
  strbuf sb;
  sb.cat (astr);
  sb.cat (authentry2str (e));

  if (!str2file (file.path, sb, perm)) {
    warn << "authadd: error writing " << file.path << "\n";
    return false;
  }

  return true;
}

bool
authadd (authentry *e) {
  /* XXX: How do we know which auth file to add to; assume the first */
  if (!authadd_core (e, userfiles[0], 0600))
    return false;

  if (shareduserfile) {
    authentry ne;
    ne.user = e->user;
    ne.pubkey = e->pubkey;
    ne.srpinfo = NULL;
    if (!authadd_core (&ne, *shareduserfile, 0644))
      return false;
  }

  return true;
}

static bool
authupdate_core (authentry *olde, authentry *newe, userfile file, int perm)
{
  authfile af;
  bool success = false;

  if (!af.openfile (file.path, O_RDONLY))
    return true;
  authentry ae;
  strbuf sb;
  while (af.getentry (&ae))
    if (olde->pubkey == ae.pubkey) {
      sb.cat (authentry2str (newe));
      success = true;
    }
    else
      sb.cat (authentry2str (&ae));
  af.closefile ();

  if (success) {
    if (!str2file (file.path, sb, perm)) {
      warn << "authupdate: error writing " << file.path << "\n";
      if (success)
	return false;
    }
    else
      success = false;
  }

  return true;
}

bool
authupdate (authentry *olde, authentry *newe)
{
  for (size_t i = 0; i < userfiles.size (); i++) {
    if (!authupdate_core (olde, newe, userfiles[i], 0600))
      return false;
  }

  if (shareduserfile) {
    authentry ne;
    ne.user = newe->user;
    ne.pubkey = newe->pubkey;
    ne.srpinfo = NULL;
    if (!authupdate_core (olde, &ne, *shareduserfile, 0644))
      return false;
  }

  return true;
}

bool
authdbinit ()
{
  authfile af;

  for (size_t i = 0; i < userfiles.size (); i++) {
    if (userfiles[i].write && !af.exists (userfiles[i].path)) {
      if (!str2file (userfiles[i].path, "", 0600)) {
	warn << "Could not create " << userfiles[i].path << "\n";
	return false;
      }
      else {
	warn << "Creating user file " << userfiles[i].path << "\n";
      }
    }
  }

  if (shareduserfile) {
    if (!af.exists (shareduserfile->path)) {
      if (!str2file (shareduserfile->path, "", 0644)) {
	warn << "Could not create " << shareduserfile->path << "\n";
	return false;
      }
      else {
	warn << "Creating user file " << shareduserfile->path << "\n";
      }
    }
  }

  return true;
}
