/* $Id: sfssesskey.C,v 1.10 1999/04/12 05:31:50 kaminsky Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfsmisc.h"
#include "crypt.h"

void
sfs_get_sesskey (sfs_hash *ksc, sfs_hash *kcs,
		 const sfs_pubkey &ks, const sfs_kmsg *smsg, 
		 const sfs_pubkey &kc, const sfs_kmsg *cmsg)
{
  sfs_sesskeydat kdat;

  kdat.type = SFS_KSC;
  kdat.ks = ks;
  kdat.sshare = smsg->ksc_share;
  kdat.kc = kc;
  kdat.cshare = cmsg->ksc_share;
  sha1_hashxdr_scrub (ksc->base (), kdat);

  kdat.type = SFS_KCS;
  kdat.sshare = smsg->kcs_share;
  kdat.cshare = cmsg->kcs_share;
  sha1_hashxdr_scrub (kcs->base (), kdat);

  bzero (&kdat.sshare, sizeof (kdat.sshare));
  bzero (&kdat.cshare, sizeof (kdat.cshare));
}

void
sfs_get_sessid (sfs_hash *sessid, const sfs_hash *ksc, const sfs_hash *kcs)
{
  sfs_sessinfo si;
  si.type = SFS_SESSINFO;
  si.ksc.set (*ksc);
  si.kcs.set (*kcs);
  sha1_hashxdr_scrub (sessid->base (), si);

  bzero (si.ksc.base (), si.ksc.size ());
  bzero (si.kcs.base (), si.kcs.size ());
}

void
sfs_get_authid (sfs_hash *authid, sfs_service service, sfs_hostname name,
		const sfs_hash *hostid, const sfs_hash *sessid)
{
  sfs_authinfo aui;
  aui.type = SFS_AUTHINFO;
  aui.service = service;
  aui.name = name;
  aui.hostid = *hostid;
  aui.sessid = *sessid;
  sha1_hashxdr (authid->base (), aui);
}

static void
set_random_key (axprt_crypt *cx, sfs_hash *sessid)
{
  sfs_hash ksc, kcs;
  rnd.getbytes (ksc.base (), ksc.size ());
  rnd.getbytes (kcs.base (), kcs.size ());
  cx->encrypt (ksc.base (), ksc.size (), kcs.base (), kcs.size ());
  if (sessid)
    sfs_get_sessid (sessid, &ksc, &kcs);
  bzero (ksc.base (), ksc.size ());
  bzero (kcs.base (), kcs.size ());
}

static inline axprt_crypt *
xprt2crypt (axprt *x)
{
  // XXX - dynamic_cast is busted in egcs
  axprt_crypt *cx = static_cast<axprt_crypt *> (&*x);
  assert (typeid (*cx) == typeid (refcounted<axprt_crypt>));
  return cx;
}

void
sfs_server_crypt (svccb *sbp, rabin_priv *sk, sfs_hash *sessid)
{
  assert (sbp->prog () == SFS_PROGRAM && sbp->proc () == SFSPROC_ENCRYPT);

  axprt_crypt *cx = xprt2crypt (sbp->getsrv ()->xprt ());
  sfs_encryptarg *arg = sbp->template getarg<sfs_encryptarg> ();
  rabin_pub kc (arg->pubkey);
  sfs_kmsg smsg;
  rnd.getbytes (&smsg, sizeof (smsg));
  sfs_encryptres res = kc.encrypt (wstr (&smsg, sizeof (smsg)));
  if (!res) {
    warn << "could not encrypt with client's public key\n";
    sbp->reject (GARBAGE_ARGS);
    set_random_key (cx, sessid);
    return;
  }
  bigint kmsg = arg->kmsg;
  sbp->reply (&res);
  
  sfs_hash ksc, kcs;
  str cmsgptxt = sk->decrypt (kmsg, sizeof (sfs_kmsg));
  if (!cmsgptxt) {
    set_random_key (cx, sessid);
    return;
  }

  sfs_get_sesskey (&ksc, &kcs, sk->n, &smsg, kc.n, sfs_get_kmsg (cmsgptxt));
  if (sessid)
    sfs_get_sessid (sessid, &ksc, &kcs);
  cx->encrypt (ksc.base (), ksc.size (), kcs.base (), kcs.size ());
  bzero (ksc.base (), ksc.size ());
  bzero (kcs.base (), kcs.size ());
  bzero (&smsg, sizeof (smsg));
  // XXX -timing attacks
}

struct sfs_client_crypt_state {
  typedef callback<void, const sfs_hash *>::ref cb_t;
  sfs_kmsg cmsg;
  sfs_encryptres res;
  ptr<axprt> x;
  ref<rabin_priv> sk;
  bigint pk;
  cb_t cb;
  sfs_client_crypt_state (ref<rabin_priv> s, const bigint &b, cb_t c)
    : sk (s), pk (b), cb (c) {}
};

static void
sfs_client_crypt_cb (sfs_client_crypt_state *st, clnt_stat err)
{
  if (err) {
    (*st->cb) (NULL);
    delete st;
    return;
  }

  sfs_hash ksc, kcs;
  str smsgptxt = st->sk->decrypt (st->res, sizeof (sfs_kmsg));
  if (!smsgptxt) {
    /* Bad ciphertext -- just start encrypting with a random key */
    rnd.getbytes (ksc.base (), ksc.size ());
    rnd.getbytes (kcs.base (), kcs.size ());
  }
  else
    sfs_get_sesskey (&ksc, &kcs, st->pk, sfs_get_kmsg (smsgptxt),
		     st->sk->n, &st->cmsg);

  sfs_hash sessid;
  sfs_get_sessid (&sessid, &ksc, &kcs);
  xprt2crypt (st->x)->encrypt (&kcs, sizeof (kcs), &ksc, sizeof (ksc));

  bzero (&ksc, sizeof (ksc));
  bzero (&kcs, sizeof (kcs));
  bzero (&st->cmsg, sizeof (st->cmsg));

  (*st->cb) (&sessid);
  delete st;
}

void
sfs_client_crypt (ptr<aclnt> c, ptr<rabin_priv> clntkey, const bigint &srvkey,
		  callback<void, const sfs_hash *>::ref cb)
{
  assert (c->rp.progno == SFS_PROGRAM && c->rp.versno == SFS_VERSION);
  sfs_client_crypt_state *st
    = New sfs_client_crypt_state (clntkey, srvkey, cb);
  st->x = c->xprt ();
  sfs_encryptarg arg;

  bzero (arg.payment.base (), arg.payment.size ()); // XXX

  rnd.getbytes (&st->cmsg, sizeof (st->cmsg));
  rabin_pub pk (srvkey);
  arg.kmsg = pk.encrypt (wstr (&st->cmsg, sizeof (st->cmsg)));

  arg.pubkey = clntkey->n;

  c->call (SFSPROC_ENCRYPT, &arg, &st->res, wrap (sfs_client_crypt_cb, st));
}

void
sfsauth_get_sesskey (sfs_hash *ksc, sfs_hash *kcs, const bigint &key)
{
  sfsauth_sesskeydat kdat;
  str raw = key.getraw ();

  kdat.type = SFS_KSC;
  kdat.key.set (const_cast<char *> (raw.cstr ()), raw.len ());
  sha1_hashxdr_scrub (ksc->base (), kdat);
  
  kdat.type = SFS_KCS;
  kdat.key.set (const_cast<char *> (raw.cstr ()), raw.len ());
  sha1_hashxdr_scrub (kcs->base (), kdat);
}

void
sfsauth_server_crypt (ptr<asrv> s, const bigint &key)
{
  sfs_hash ksc, kcs;
  sfsauth_get_sesskey (&ksc, &kcs, key);

  axprt_crypt *cx = xprt2crypt (s->xprt ());
  cx->encrypt (ksc.base (), ksc.size (), kcs.base (), kcs.size ());

  bzero (ksc.base (), ksc.size ());
  bzero (kcs.base (), kcs.size ());
}

void
sfsauth_client_crypt (ptr<aclnt> c, const bigint &key)
{
  sfs_hash ksc, kcs;
  sfsauth_get_sesskey (&ksc, &kcs, key);

  axprt_crypt *cx = xprt2crypt (c->xprt ());
  cx->encrypt (kcs.base (), kcs.size (), ksc.base (), ksc.size ());

  bzero (ksc.base (), ksc.size ());
  bzero (kcs.base (), kcs.size ());
}
