// -*-c++-*-
/* $Id: sfsrwcd.h,v 1.36 1999/05/17 03:33:11 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "arpc.h"
#include "sfscd_prot.h"
#include "nfstrans.h"
#include "sfsmisc.h"
#ifdef USE_UVFS
#include "uvfstrans.h"
#endif /* USE_UVFS */
#include "qhash.h"
#include "itree.h"
#include "crypt.h"
#include "list.h"

#ifdef USE_UVFS
extern bool uvfs;
#else /* !USE_UVFS */
enum { uvfs = 0 };
#endif /* !USE_UVFS */

extern AUTH *aui0;
extern blowfish fhkey;

template<> struct qhash_lookup_return<auto_auth> {
  static auto_auth nullauth;
  typedef auto_auth &type;
  typedef const auto_auth &const_type;
  static type ret (auto_auth *a) { if (a) return *a; return nullauth = NULL; }
};

class attr_cache {
  struct access_dat {
    u_int32_t mask;
    u_int32_t perm;

    access_dat (u_int32_t m, u_int32_t p) : mask (m), perm (p) {}
  };

public:
  struct attr_dat {
    attr_cache *const cache;
    const nfs_fh3 fh;
    ex_fattr3 attr;
    qhash<sfs_aid, access_dat> access;

    ihash_entry<attr_dat> fhlink;
    tailq_entry<attr_dat> lrulink;

    attr_dat (attr_cache *c, const nfs_fh3 &f, const ex_fattr3 *a);
    ~attr_dat ();
    void touch ();
    void set (const ex_fattr3 *a, const wcc_attr *w);
    bool valid () { return timenow < implicit_cast<time_t> (attr.expire); }
  };

private:
  friend class attr_dat;
  ihash<const nfs_fh3, attr_dat, &attr_dat::fh, &attr_dat::fhlink> attrs;

  static void remove_aid (sfs_aid aid, attr_dat *ad)
    { ad->access.remove (aid); }

public:
  ~attr_cache () { attrs.deleteall (); }
  void flush_attr () { attrs.deleteall (); }
  void flush_access (sfs_aid aid) { attrs.traverse (wrap (remove_aid, aid)); }
  void flush_access (const nfs_fh3 &fh, sfs_aid);

  void attr_enter (const nfs_fh3 &, const ex_fattr3 *, const wcc_attr *);
  const ex_fattr3 *attr_lookup (const nfs_fh3 &);

  void access_enter (const nfs_fh3 &, sfs_aid aid,
		     u_int32_t mask, u_int32_t perm);
  int32_t access_lookup (const nfs_fh3 &, sfs_aid, u_int32_t mask);
};

struct server;
struct userauth : public virtual refcount {
public:
  const sfs_aid aid;
  const ref<server> sp;

private:
  vector<svccb *> sbpvec;
  timecb_t *tmo;
  callbase *cbase;
  bool aborted;

  sfsagent_auth_res ares;
  sfs_loginres sres;

  sfs_seqno seqno;

  int ntries;

  void finish (u_int32_t authno);
  void aresult (clnt_stat);
  void sresult (clnt_stat);
  void timeout ();
  void sendreq ();

public:
  userauth (sfs_aid aid, const ref<server> &s);
  ~userauth ();
  void pushreq (svccb *sbp);
  void abort ();
};

class server : public virtual refcount {
private:
  static u_int32_t srvno_next;
  static u_int32_t gensrvno ();

  qhash<sfs_aid, auto_auth> authnos;
  qhash<sfs_aid, ref<userauth> > authpending;
  vector<svccb *> waitq;
  timecb_t *recon_tmo;
  int recon_backoff;
  bool lock_flag;
  bool condemned;

  attr_cache ac;

  static bool getserver (ptr<server> *sp, u_int32_t *, svccb *sbp);
  static void getcall (svccb *sbp);
  static void doflushuser (sfs_aid aid, server *sp);

  bool authok (svccb *sbp);
  void nfs3dispatch (svccb *);
#ifdef USE_UVFS
  void uvfsdispatch (svccb *);
  void uvfsreaddir (svccb *);
#endif /* USE_UVFS */
  void getreply (time_t rqtime, svccb *sbp, void *res, clnt_stat err);
  void reconnect ();
  void reconnectcb (const sfs_fsinfo *, int err);
  void condemn ();

protected:
  server (const str &path, const sfs_hash &hostid,
	  const sfs_hostinfo &hostinfo);
  ~server ();

public:
  const u_int32_t srvno;
  const str path;
  const sfs_hash hostid;
  const sfs_hostinfo hostinfo;

  ptr<axprt_crypt> x;
  ptr<aclnt> sfsc;
  ptr<aclnt> nfsc;
  ptr<asrv> nfscbs;
  ptr<asrv> sfscbs;

  sfs_fsinfo fsinfo;
  sfs_authinfo authinfo;
  sfs_seqno seqno;

  ihash_entry<server> pathlink;
  ihash_entry<server> srvnolink;

  static void init ();
  static void alloc (svccb *, int fd);
  static server *lookup (const str &path);
  static void flushuser (sfs_aid aid);
  static void sendreply (svccb *sbp, void *res, u_int32_t srvno);

  void dispatch (svccb *sbp);
  void cbdispatch (svccb *sbp);

  bool restart (svccb *sbp, clnt_stat err);

  void lock () { assert (!lock_flag); lock_flag = true; }
  void unlock ();
  bool locked () { return lock_flag; }
  void destroy ();

  void mapuser (sfs_aid aid, u_int32_t authno);
  void clear ()
    { authnos.clear (); authpending.clear (); ac.flush_attr (); }
  AUTH *getauth (sfs_aid aid) { return authnos[aid]; }
};

extern ihash<const u_int32_t, server,
  &server::srvno, &server::srvnolink> &srvnotab;

class srvcon {
  typedef callback<void, const sfs_fsinfo *, int>::ref cb_t;

  const ref<server> s;
  const cb_t cb;
  const ref<rabin_priv> kc;

  sfs_connectres conres;
  sfs_fsinfo fsires;

  void fail (int err);
  void setfd (int fd);

  void getsock () {
    tcpconnect (s->hostinfo.hostname, sfs_port,
		wrap (this, &srvcon::getsockres));
  }
  void getsockres (int fd);
  void sfsconres (clnt_stat);
  void crypt () {
    sfs_client_crypt (s->sfsc, kc, s->hostinfo.pubkey,
		      wrap (this, &srvcon::cryptres));
  }
  void cryptres (const sfs_hash *sessid);
  void getfsinfo () {
    s->sfsc->call (SFSPROC_GETFSINFO, NULL, &fsires,
		   wrap (this, &srvcon::getfsinfores));
  }
  void getfsinfores (clnt_stat);

public:
  srvcon (const ref<server> &s, int fd, cb_t cb);
};

extern ptr<rabin_priv> client_key;
extern ptr<axprt_unix> cdx;
extern ptr<aclnt> cdc;

void sfsctl_accept (ptr<axprt_stream> x, const authunix_parms *aup);
bool sfsctl_intercept (server *s, u_int32_t srvno, svccb *sbp);
