/* $Id: filesrv.C,v 1.22 1999/05/17 08:29:54 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfsrwsd.h"

#define FSID_SHIFT 20

static u_int64_t
fattr2ino (const fattr3 &a)
{
  return a.fileid ^ a.fsid << FSID_SHIFT;
}

int
filesrv::path2fsidx (str path, size_t nfs)
{
  int best = 0;
  for (size_t i = 1; i < nfs; i++) {
    const str &iname = fstab[i].path_mntpt;
    if (iname.len () > fstab[best].path_mntpt.len ()
	&& iname.len () < path.len ()
	&& path[iname.len ()] == '/'
	&& !memcmp (iname, path, iname.len ()))
      best = i;
  }
  return best;
}

filesrv::filesrv ()
  : st (synctab_alloc ())
{
}

void
filesrv::init (cb_t c)
{
  assert (!cb);
  cb = c;
  aclntudp_create (host, 0, nfs_program_3, wrap (this, &filesrv::getnfsc));
}

void
filesrv::getnfsc (ptr<aclnt> nc, clnt_stat stat)
{
  if (!nc) {
    warn << host << ": NFS3: " << stat << "\n";
    (*cb) (false);
    return;
  }
  c = nc;
  aclntudp_create (host, 0, mount_program_3, wrap (this, &filesrv::getmountc));
}

void
filesrv::getmountc (ptr<aclnt> nc, clnt_stat stat)
{
  if (!nc) {
    warn << host << ": MOUNT3: " << stat << "\n";
    (*cb) (false);
    return;
  }
  mountc = nc;

  fstab[0].parent = &fstab[0];
  if (fstab[0].path_root)
    getfh3 (mountc, fstab[0].path_root, wrap (this, &filesrv::gotroot, 0));
  else
    gotroot (0, &fstab[0].fh_root, MNT3_OK, RPC_SUCCESS);
}

void
filesrv::gotroot (size_t fsno, const nfs_fh3 *fhp,
		  mountstat3 mstat, clnt_stat rstat)
{
  str path = fstab[fsno].path_root;
  if (rstat) {
    warn << "getfh (" << path << "): " << rstat << "\n";
    (*cb) (false);
    return;
  }
  else if (mstat) {
    warn << "getfh (" << path << "): " << mstat << "\n";
    (*cb) (false);
    return;
  }

  filesys &fs = fstab[fsno];

  fs.fh_root = *fhp;
  if (!fsno) {
    fs.fh_mntpt = *fhp;
    if (fstab.size () <= 1)
      getattr ();
    else if (fstab[1].path_root)
      getfh3 (mountc, fstab[1].path_root, wrap (this, &filesrv::gotroot, 1));
    else
      gotroot (1, &fstab[1].fh_root, MNT3_OK, RPC_SUCCESS);
    return;
  }

  filesys *mfs = &fstab[path2fsidx (fstab[fsno].path_mntpt, fsno)];
  fstab[fsno].parent = mfs;
  lookupfh3 (c, mfs->fh_root,
	     substr (fs.path_mntpt, mfs->path_mntpt.len ()),
	     wrap (this, &filesrv::gotmntpt, fsno));
}

void
filesrv::gotmntpt (size_t fsno, const nfs_fh3 *fhp,
		   nfsstat3 nstat, clnt_stat rstat)
{
  str path = fstab[fsno].path_mntpt;
  if (rstat) {
    warn << "lookup (" << path << "): " << rstat << "\n";
    (*cb) (false);
    return;
  }
  else if (nstat) {
    warn << "lookup (" << path << "): " << nstat << "\n";
    (*cb) (false);
    return;
  }

  filesys &fs = fstab[fsno];
  fs.fh_mntpt = *fhp;
  if (fsno + 1 >= fstab.size ())
    getattr ();
  else if (fstab[fsno + 1].path_root)
    getfh3 (host, fstab[fsno + 1].path_root,
	    // XXX - int (...) is for g++ bug
	    wrap (this, &filesrv::gotroot, int (fsno + 1)));
  else
    gotroot (fsno + 1, &fstab[fsno + 1].fh_root,
	     MNT3_OK, RPC_SUCCESS);
}

static fattr3 &
gfsno2fattr (filesys &fs, size_t gfsno)
{
  switch (gfsno & 3) {
  case 0:
    return fs.attr_root;
  case 1:
    return fs.attr_mntpt;
  case 2:
    return fs.attr_root_dd;
  default:			// XXX - egcs
    return fs.attr_mntpt_dd;
  }
}


void
filesrv::gotattr (getattr_state *sp, size_t fsno, getattr3res *resp,
		  clnt_stat err)
{
  filesys *fsp = &fstab[fsno >> 2];
  str path = fsno & 1 ? fsp->path_mntpt : fsp->path_root;
  if (err) {
    warn << path << ": " << err << "\n";
    sp->ok = false;
  }
  else if (resp->status) {
    warn << path << ": " << resp->status << "\n";
    sp->ok = false;
  }
  else
    gfsno2fattr (*fsp, fsno) = *resp->attributes;
  delete resp;

  if (!--sp->nleft) {
    if (sp->ok)
      finish ();
    else 
      (*cb) (false);
  }
}

void
filesrv::gotdd (getattr_state *sp, size_t fsno, lookup3res *resp,
		clnt_stat err)
{
  filesys *fsp = &fstab[fsno >> 2];
  str path = fsno & 1 ? fsp->path_mntpt : fsp->path_root;
  if (err) {
    warn << path << "/..: " << err << "\n";
    sp->ok = false;
  }
  else if (resp->status) {
    warn << path << "/..: " << resp->status << "\n";
    sp->ok = false;
  }
  else if (resp->resok->obj_attributes.present)
    gfsno2fattr (*fsp, fsno) = *resp->resok->obj_attributes.attributes;
  else {
    getattr3res *nresp = New getattr3res;
    c->call (NFSPROC3_GETATTR, &resp->resok->object, nresp,
	     wrap (this, &filesrv::gotattr, sp, fsno, nresp), auth_root);
    delete resp;
    return;
  }
  delete resp;

  if (!--sp->nleft) {
    if (sp->ok)
      finish ();
    else 
      (*cb) (false);
  }
}

void
filesrv::getattr ()
{
  getattr_state *sp = New getattr_state;
  sp->nleft = 4 * fstab.size () - 2;
  sp->ok = true;
  for (size_t i = 0, n = sp->nleft + 2; i < n; i++) {
    if (i == 1 || i == 3)
      continue;
    filesys *fsp = &fstab[i >> 2];
    if (!(i & 2)) {
      getattr3res *resp = New getattr3res;
      c->call (NFSPROC3_GETATTR,
	       (i&1) ? &fsp->fh_mntpt : &fsp->fh_root,
	       resp, wrap (this, &filesrv::gotattr, sp, i, resp), auth_root);
    }
    else {
      lookup3res *resp = New lookup3res;
      diropargs3 arg;
      arg.dir = (i&1) ? fsp->fh_mntpt : fsp->fh_root;
      arg.name = "..";
      c->call (NFSPROC3_LOOKUP, &arg, resp,
	       wrap (this, &filesrv::gotdd, sp, i, resp), auth_root);
    }
  }
}

void
filesrv::finish ()
{
  str p;
  sha1ctx sc;
  p = str2wstr (sk->p.getraw ());
  sc.update (p, p.len ());
  sc.update (hostid.base (), hostid.size ());
  p = str2wstr (sk->q.getraw ());
  sc.update (p, p.len ());
  char key[sha1::hashsize];
  sc.final (key);
  fhkey.setkey (key, sizeof (key));
  bzero (key, sizeof (key));

  fstab[0].attr_mntpt = fstab[0].attr_root;
  fstab[0].attr_mntpt.fileid = 1;
  fstab[0].attr_mntpt_dd = fstab[0].attr_root;
  fstab[0].attr_mntpt_dd.fileid = 1;

  for (filesys *fsp = fstab.base (); fsp < fstab.lim (); fsp++) {
    mp3tab.insert (fsp);
    root3tab.insert (fsp);
    fsp->inotab = New filesys::inotab_t;
    fsp->parent->inotab->insert (fsp->attr_mntpt.fileid,
				 fattr2ino (fsp->attr_root));
    fsp->inotab->insert (fsp->attr_root_dd.fileid,
			 fattr2ino (fsp->attr_mntpt_dd));
  }

#if 0
  for (filesys *fsp = fstab.base (); fsp < fstab.lim (); fsp++) {
    warnx << fsp->path_mntpt << "\n"
	  << "     mntpt: "
	  << strbuf ("%qx", fsp->attr_mntpt.fsid) << " "
	  << strbuf ("%qx", fsp->attr_mntpt.fileid) << "\n"
	  << "  mntpt/..: "
	  << strbuf ("%qx", fsp->attr_mntpt_dd.fsid) << " "
	  << strbuf ("%qx", fsp->attr_mntpt_dd.fileid) << "\n"
	  << "      root: "
	  << strbuf ("%qx", fsp->attr_root.fsid) << " "
	  << strbuf ("%qx", fsp->attr_root.fileid) << "\n"
	  << "   root/..: "
	  << strbuf ("%qx", fsp->attr_root_dd.fsid) << " "
	  << strbuf ("%qx", fsp->attr_root_dd.fileid) << "\n";
  }
#endif

  fh3trans fht (fh3trans::ENCODE, fhkey, 0);

  fsinfo.set_prog (ex_NFS_PROGRAM);
  fsinfo.nfs->set_vers (ex_NFS_V3);
  fsinfo.nfs->v3->root = fstab[0].fh_root;
  if (!rpc_traverse (fht, fsinfo.nfs->v3->root))
    fatal ("filesrv::finish: nfs3_transres encode failed (err %d)\n", fht.err);

  fsinfo.nfs->v3->subfs.setsize (fstab.size () - 1);
  for (size_t i = 1; i < fstab.size (); i++) {
    fsinfo.nfs->v3->subfs[i-1].path = fstab[i].path_mntpt;
    fsinfo.nfs->v3->subfs[i-1].fh = fstab[i].fh_root;
    fht.srvno = i;
    if (!rpc_traverse (fht, fsinfo.nfs->v3->subfs[i-1].fh))
      fatal ("filesrv::finish: nfs3_transres encode failed (err %d)\n",
	     fht.err);
  }

  mountc = NULL;
  (*cb) (true);
  cb = NULL;
}

bool
filesrv::getauthclnt ()
{
  if (authxprt && !authxprt->ateof ())
    return true;
  int fd = suidgetfd ("authserv");
  if (fd < 0) {
    authxprt = NULL;
    authclnt = NULL;
    return false;
  }
  authclnt = aclnt::alloc (authxprt = axprt_stream::alloc (fd),
			   sfsauth_program_1);
  return true;
}

int
filesrv::fhsubst (bool *substp, nfs_fh3 *fhp, u_int32_t *fhnop)
{
  filesys *fsp;
  for (fsp = mp3tab[*fhp]; fsp; fsp = mp3tab.nextkeq (fsp))
    if (fsp != fstab.base () && getfsno (fsp->parent) == *fhnop) {
      *fhp = fsp->fh_root;
      *fhnop = fsp - fstab.base ();
      *substp = true;
      break;
    }
  return 0;
}

static inline bool
anon_checkperm (svccb *sbp, u_int options, bool isroot)
{
  switch (sbp->proc ()) {
  case NFSPROC3_GETATTR:
  case NFSPROC3_ACCESS:
  case NFSPROC3_FSINFO:
  case NFSPROC3_PATHCONF:
    if (isroot || (options & filesys::ANON_READ) == filesys::ANON_READ)
      return true;
    break;

  case NFSPROC3_FSSTAT:
    {
      if (isroot || (options & filesys::ANON_READ) == filesys::ANON_READ)
	return true;
      fsstat3res res (NFS3_OK);
      res.resok->tbytes = 0;
      res.resok->fbytes = 0;
      res.resok->abytes = 0;
      res.resok->tfiles = 0;
      res.resok->ffiles = 0;
      res.resok->afiles = 0;
      res.resok->invarsec = 0;
      sbp->reply (&res);
      return false;
    }

  case NFSPROC3_LOOKUP:
  case NFSPROC3_READLINK:
  case NFSPROC3_READ:
  case NFSPROC3_READDIR:
  case NFSPROC3_READDIRPLUS:
    if (options & filesys::ANON_READ)
      return true;
    break;

  default:
    if ((options & filesys::ANON_READWRITE) == filesys::ANON_READWRITE)
      return true;
    break;
  }

  nfs3exp_err (sbp, NFS3ERR_PERM);
  return false;
}

bool
filesrv::fixarg (svccb *sbp, reqstate *rqsp)
{
  fh3trans fht (fh3trans::DECODE, fhkey);
  if (!nfs3_transarg (fht, sbp->template getarg<void> (), sbp->proc ())) {
    nfs3exp_err (sbp, nfsstat3 (fht.err));
    return false;
  }
  if (fht.srvno >= fstab.size ()) {
    nfs3exp_err (sbp, NFS3ERR_BADHANDLE);
    return false;
  }
  rqsp->fsno = fht.srvno;
  filesys *fsp = &fstab[rqsp->fsno];
  rqsp->rootfh = false;

  /* We let anonymous users GETATTR any root file handle, not just the
   * root of all exported files.  This is to help client
   * implementations that want to avoid (st_ino, st_dev) conflicts by
   * creating multiple mount points for each server.  Is this bad? */
  if (!sbp->getaui ()
      && !anon_checkperm (sbp, fsp->options,
			  *sbp->template getarg<nfs_fh3> () == fsp->fh_root))
    return false;

  switch (sbp->proc ()) {
  case NFSPROC3_LOOKUP:
    {
      diropargs3 *doa = sbp->template getarg<diropargs3> ();
      if (doa->name == ".." && doa->dir == fsp->fh_root) {
	if (!getfsno (fsp)) {
	  nfs3exp_err (sbp, NFS3ERR_ACCES);
	  return false;
	}
	doa->dir = fsp->fh_mntpt;
	rqsp->fsno = getfsno (fsp->parent);
	rqsp->rootfh = true;
      }
      break;
    }
  case NFSPROC3_READDIR:
  case NFSPROC3_READDIRPLUS:
    {
      nfs_fh3 *rpa = sbp->template getarg<nfs_fh3> ();
      if (*rpa == fsp->fh_root)
	rqsp->rootfh = true;
      break;
    }
  }
  return true;
}

void
filesrv::fixrdres (void *_res, filesys *fsp, bool rootfh)
{
  readdir3res *res = static_cast<readdir3res *> (_res);
  if (res->status)
    return;
  for (entry3 *e = res->resok->reply.entries; e; e = e->nextentry) {
    u_int64_t *idp = (*fsp->inotab)[e->fileid];
    if (idp && e->name != "." && (rootfh || e->name != ".."))
      e->fileid = *idp;
    else
      e->fileid ^= fsp->attr_root.fsid << FSID_SHIFT;
  }
}

void
filesrv::fixrdplusres (void *_res, filesys *fsp, bool rootfh)
{
  readdirplus3res *res = static_cast<readdirplus3res *> (_res);
  if (res->status)
    return;
  for (entryplus3 *e = res->resok->reply.entries; e; e = e->nextentry) {
    u_int64_t *idp = (*fsp->inotab)[e->fileid];
    if (idp && e->name != "." && (rootfh || e->name != ".."))
      e->fileid = *idp;
    else
      e->fileid ^= fsp->attr_root.fsid << FSID_SHIFT;
    if (!e->name_handle.present)
      e->name_attributes.set_present (false);
    else if (rootfh && e->name == "..") {
      e->name_attributes.set_present (false);
      e->name_handle.set_present (false);
    }
    else {
      filesys *nfsp;
      for (nfsp = mp3tab[*e->name_handle.handle];
	   nfsp && nfsp->parent != fsp; nfsp = mp3tab.nextkeq (nfsp))
	;
      if (nfsp)
	e->name_attributes.set_present (false);
    }
  }
}

static int
xor_ino (fattr3 *fp, u_int32_t srvno)
{
  fp->fileid = fattr2ino (*fp);
  return 0;
}
static fh3trans::fattr_hook_t fhook = wrap (xor_ino);

bool
filesrv::fixres (svccb *sbp, void *res, reqstate *rqsp)
{
  filesys *fsp = &fstab[rqsp->fsno];

  if (sbp->proc () == NFSPROC3_READDIR)
    fixrdres (res, fsp, rqsp->rootfh);
  else if (sbp->proc () == NFSPROC3_READDIRPLUS)
    fixrdplusres (res, fsp, rqsp->rootfh);

  bool subst = false;
  fh3trans fht (fh3trans::ENCODE, fhkey, rqsp->fsno,
		wrap (this, &filesrv::fhsubst, &subst));

  /* XXX - the next line is to make the leases sort of work */
  nfs3_transarg (fht, sbp->getvoidarg (), sbp->proc ());

  fht.fattr_hook = fhook;
  if (!nfs3_transres (fht, res, sbp->proc ())) {
    warn ("nfs3reply: nfs3_transres encode failed (err %d)\n", fht.err);
    nfs3exp_err (sbp, nfsstat3 (fht.err));
    return false;
  }

  switch (sbp->proc ()) {
  case NFSPROC3_LOOKUP:
    {
      lookup3res *dor = static_cast<lookup3res *> (res);
      if (rqsp->rootfh) {
	if (dor->status)
	  dor->resfail->set_present (false);
	else
	  dor->resok->dir_attributes.set_present (false);
      }
      if (subst && !dor->status)
	dor->resok->obj_attributes.set_present (false);
      break;
    }
  case NFSPROC3_ACCESS:
    if (!sbp->getaui () && fsp->options != filesys::ANON_READWRITE) {
      access3res *ar = static_cast<access3res *> (res);
      if ((fsp->options & filesys::ANON_READ) != filesys::ANON_READ) {
	if (ar->status)
	  ar->resfail->set_present (false);
	else {
	  ar->resok->access = 0;
	  ar->resok->obj_attributes.set_present (false);
	}
      }
      else if (!ar->status)
	ar->resok->access &= ACCESS3_READ|ACCESS3_LOOKUP|ACCESS3_EXECUTE;
    }
    break;
  }
  return true;
}
