// -*- c++ -*-
/* $Id: sfsrwsd.h,v 1.34 1999/05/17 06:54:50 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "qhash.h"
#include "arpc.h"
#include "vector.h"
#include "sfsmisc.h"
#include "mount_prot.h"
#include "crypt.h"
#include "rabin.h"
#include "seqno.h"
#include "nfstrans.h"

inline bool
operator== (const nfs_fh3 &a, const nfs_fh3 &b)
{
  return a.data.size () == b.data.size ()
    && !memcmp (a.data.base (), b.data.base (), b.data.size ());
}
inline bool
operator!= (const nfs_fh3 &a, const nfs_fh3 &b)
{
  return !(a == b);
}

struct hashfh3 {
  hashfh3 () {}
  hash_t operator() (const nfs_fh3 &fh) const {
    const u_int32_t *s = reinterpret_cast<const u_int32_t *> (fh.data.base ());
    const u_int32_t *e = s + (fh.data.size () >> 2);
    u_int32_t val = 0;
    while (s < e)
      val ^= *s++;
    return val;
  }
};

struct filesys {
  typedef qhash<u_int64_t, u_int64_t> inotab_t;

  filesys *parent;
  str path_root;		// Local path corresponding to root
  str path_mntpt;		// Mountpoint relative to exported namespace
  nfs_fh3 fh_root;		// NFS File handle of root
  nfs_fh3 fh_mntpt;		// NFS File handle of mount point
  fattr3 attr_root;		// Attributes of root
  fattr3 attr_mntpt;		// Attributes of mount point
  fattr3 attr_root_dd;		// Attributes of root/..
  fattr3 attr_mntpt_dd;		// Attributes of mount point/..
  enum {
    ANON_READ = 1,
    ANON_READWRITE = 3,
  };
  u_int options;		// Any of the above options
  ihash_entry<filesys> rhl;
  ihash_entry<filesys> mphl;
  inotab_t *inotab;
};

struct synctab;
class filesrv {
public:
  struct reqstate {
    u_int32_t fsno;
    bool rootfh;
  };

  str host;
  ptr<aclnt> c;
  ptr<aclnt> mountc;
  
  sfs_hostinfo hinfo;
  sfs_hash hostid;
  ptr<rabin_priv> sk;

  ptr<axprt_stream> authxprt;
  ptr<aclnt> authclnt;

  vector<filesys> fstab;
  ihash<nfs_fh3, filesys, &filesys::fh_root, &filesys::rhl, hashfh3> root3tab;
  ihash<nfs_fh3, filesys, &filesys::fh_mntpt, &filesys::mphl, hashfh3> mp3tab;

  blowfish fhkey;
  sfs_fsinfo fsinfo;

private:
  typedef callback<void, bool>::ref cb_t;
  cb_t::ptr cb;
  struct getattr_state {
    int nleft;
    bool ok;
  };

  PRIVDEST ~filesrv ();		// No deleting

  /* Initialization functions */
  int path2fsidx (str path, size_t nfs);
  void getnfsc (ptr<aclnt> c, clnt_stat stat);
  void getmountc (ptr<aclnt> c, clnt_stat stat);
  void gotroot (size_t fsno, const nfs_fh3 *fhp,
		mountstat3 mstat, clnt_stat rstat);
  void gotmntpt (size_t fsno, const nfs_fh3 *fhp,
		 nfsstat3 nstat, clnt_stat rstat);
  void fixrdres (void *res, filesys *fsp, bool rootfh);
  void fixrdplusres (void *res, filesys *fsp, bool rootfh);
  void getattr ();
  void gotattr (getattr_state *, size_t fsno, getattr3res *, clnt_stat);
  void gotdd (getattr_state *, size_t fsno, lookup3res *, clnt_stat);
  void finish ();

  int fhsubst (bool *substp, nfs_fh3 *fhp, u_int32_t *fsnop);
  size_t getfsno (const filesys *fsp) const {
#ifdef CHECK_BOUNDS
    assert (fstab.base () <= fsp && fsp < fstab.lim ());
#endif CHECK_BOUNDS
    return fsp - fstab.base ();
  }

public:
  synctab *const st;

  void init (cb_t cb);

  bool fixarg (svccb *sbp, reqstate *fsnop);
  bool fixres (svccb *sbp, void *res, reqstate *fsnop);

  bool getauthclnt ();

  filesrv ();
};

extern int sfssfd;

extern AUTH *auth_root;

const strbuf &strbuf_cat (const strbuf &, mountstat3);
void getfh3 (const char *host, str path,
	     callback<void, const nfs_fh3 *, mountstat3, clnt_stat>::ref);
void getfh3 (ref<aclnt> c, str path,
	     callback<void, const nfs_fh3 *, mountstat3, clnt_stat>::ref);
void lookupfh3 (ref<aclnt> c, const nfs_fh3 &start, str path,
		callback<void, const nfs_fh3 *, nfsstat3, clnt_stat>::ref cb);

class client : public virtual refcount {
  filesrv *fsrv;

  vector<auto_auth> authtab;
  vector<sfs_cred> credtab;
  vector<size_t> authfreelist;

  ptr<axprt_crypt> sfsx;
  ptr<asrv> nfssrv;
  ptr<asrv> sfssrv;

  sfs_hash sessid;
  sfs_hash authid;
  seqcheck seqstate;
  bool authid_valid;

  static u_int64_t nextgen ();

  void fail ();
  size_t authalloc ();
  void authfree (size_t);

  void dologin (svccb *);
  void dologin_cb (svccb *, sfsauth_loginres *, clnt_stat);
  void doidnames (svccb *);
  void doidnums (svccb *);

  void nfs3dispatch (svccb *);
  void nfs3reply (svccb *sbp, void *res, filesrv::reqstate rqs, clnt_stat err);
  void sfsdispatch (svccb *);

protected:
  client (ref<axprt_crypt> x);
  ~client ();

public:
  ptr<aclnt> nfscbc;
  const u_int64_t generation;
  ihash_entry<client> glink;

  static void launch (ref<axprt_crypt> x) { vNew refcounted<client> (x); }
  filesrv *getfilesrv () const { return fsrv; }
};

extern ihash<const u_int64_t, client,
  &client::generation, &client::glink> clienttab;

synctab *synctab_alloc ();
void synctab_free (synctab *st);
void doleases (client *c, svccb *sbp, void *res);

bool fh3tosfs (nfs_fh3 *);
bool fh3tonfs (nfs_fh3 *);

ptr<axprt_stream> client_accept (int fd);
void client_accept_standalone ();

extern filesrv *defsrv;
